;-----------------------------------------------------------------------------
;       Copyright (C) 2002 Drawbridge Swamp                 
; 	All rights reserved.
;
;
;       FILE NAME       :       EMF.ASM 
; 	TARGET MCU	:	C8051F300 
;       DESCRIPTION     :       DCC decoder                 
;       Status: 10/16/2006      EMF implemented, most functions 
;                               tested
;                               EMF algorithm by Gil Fuchs/Mike Bolton, Kfr skipped
;                               Acc/Dec now OK 28/127 speed steps 
;                               Second address added for steering of bus
;                               even = motor+functions,next odd steering
;				CVStorage changed to 01A00x (diesel CPU problem
;  direct write into RAM OK
;-----------------------------------------------------------------------------


;-----------------------------------------------------------------------------
; EQUATES
;-----------------------------------------------------------------------------

FREQUENCY       EQU     45      ; Oscillator in F302
BUS     EQU     0       ; don't compile Bus
;BUS     EQU     1       ; compile Bus

$include (c8051f300.inc)					; Include register definition file.

MODADD          EQU	6
ERRMOD          EQU	72
ENDMOD	        EQU	78
SYNCP	        EQU	10
RINGI_1         EQU     10
RINGI_2         EQU     3
RINGI_3         EQU     0

GTAKT           EQU     2
BREMS2          EQU     2

;GREEN_LED			EQU		P0.2
FUN1            EQU     P0.7
FUN2            EQU     P0.7
FUN3            EQU     P0.7
FUN4            EQU     P0.7
FUN5            EQU     P0.7
FUN6            EQU     P0.7
FUN7            EQU     P0.7
FUN8            EQU     P0.7

IF BUS
FLIGHT	        EQU	P0.7
ELSE
FLIGHT          EQU     P0.1
ENDIF

RLIGHT	        EQU	P0.7

PULSIN	        EQU	P0.0           ; Port I/O pin connected to Green LED.

IF BUS
PWMOUT          EQU     P0.1            ; 
ENDIF


Rmode 	        EQU	R1
Rtimbas	        EQU	R2
Rtimbuf	        EQU	R3
Rbytec	        EQU	R4
Rsync	        EQU	R5
Rmode_1	        EQU	9
Rtimbas_1       EQU     10							; Mode register, when in bank 0
Rbytec_1        EQU	12
Rsync_1	        EQU	13


FORWARD		EQU	flags.1
OldBit		EQU	flags.4
ADReady         EQU     flags.5
SERVMOD         EQU	flags.6         ; service mode on/off
STOPPED         EQU     flags.7
CONSIST_DECODED	EQU	flags1.4  
CONSIST_ACTIVE	EQU     flags1.6  
FRONT_LIGHT	EQU	CV22.7
REAR_LIGHT	EQU	CV22.6
LIGHT_BASE	EQU	CV22.4
LIGHT_CONSIST	EQU	CV22.5


;CVStorage       EQU     01C00h          ; begin of CV storage area in flash eeprom
CVStorage       EQU     01A00h          ; Diesel CPU has a problem with 01C00h
	


;-----------------------------------------------------------------------------
; VARIABLES
;-----------------------------------------------------------------------------

;-------------------
; STACK

STACK           SEGMENT IDATA        ; declare STACK segment
                RSEG  STACK
                DS    80h            ; reserve 128 bytes for stack

;-----------------------------------------------------------------------------
; MACRO DEFINITIONS
;-----------------------------------------------------------------------------
;-----------------------------------------------------------------------------
; VARIABLES
;-----------------------------------------------------------------------------


                BSEG

                org 0h

 

                DSEG at 20h

flags:  DS      1	;  0-timer overflow/ 1-actual reverse
		; 2-set direction/ 3-blink
		; 4-saved pulse level/ 5-EMF done
		; 6-service/ 7 emf-NC 
				
flags1:	        DS	1       ; 0-3byte message,1-4byte
                                ; 2-5byte,3-6byte 
                                ; 4-Consist_decoded,5-allowed function,
                                ; 6-Consist_active, 7-Remfun
                                ; >>> 0ed @prepint !!!
byte_1:	        DS	1
byte_2:	        DS	1
byte_3:	        DS	1
byte_4:	        DS	1
byte_5:	        DS	1
byte_6:	        DS	1
lokfun:         DS      1
strps:          DS      1
CV22:	        DS	1
CV29:	        DS	1
CV17:   	DS	1
CV18:   	DS	1
CV19:   	DS	1
CV21:   	DS	1
; end of bit-adressable section
CV1:    	DS	1
CV2:    	DS	1
CV3:    	DS	1
CV4:    	DS	1
CV5:    	DS	1
CV8:    	DS	1
CV9:    	DS	1
EMF_Kp:
CV52:   	DS	1
EMF_Ki:
CV53:   	DS	1
EMF_Kfr:
CV54:   	DS	1
Creep:
CV55:           DS      1       ; Creep Speed if DCC Signal lost
CV67:           DS      28      ; table entries
lokist:         DS	1
loksoll:        DS      1
duty:           DS      1
IF BUS
duty1:          DS      1       ; PWM for servo
period:         DS      1       ; counter for servo PWM overflows to 256
pwmtim:         DS      1       ; PWM timer
steerlow:       DS      1
steerhgh:       DS      1
ENDIF
reload:         DS      1
failsf:         DS      1

x_base:	        DS	1
x_value:        DS	1 

l_time:
                DS    1     ; Timer reload values low byte
timer1:	        DS	1
timer2:	        DS	1
scratch:        DS      1
lasxor:         DS      1
pagreg:         DS      1
timacc:         DS      1
PWM_val:        DS      1
;EMF_Kp:         DS      1
;EMF_Ki:         DS      1				
;EMF_Kfr:        DS      1
Integ:          DS      1
EEADR:          DS      1

;-----------------------------------------------------------------------------
; MAIN PROGRAM CODE
;-----------------------------------------------------------------------------

        CSEG AT 0

        ljmp  begin                    ; RESET vector


;EMF01      SEGMENT  CODE                 ; declare CODE segment
;            RSEG  EMF01                  ; select CODE segment
;            USING 0                       ; using register bank 0




; 264 cycles in interrupt @24MHz ^ 11usec

	CSEG AT 0bh

;Rmode,Rtimbas,Flags.Oldbit,Rtimbuf,Rsync

BitCollect:
;					setb		GREEN_LED
        push    ACC
	push	DPH
	push	DPL
        push    PSW
        mov     PSW,#00001000B          ; use register bank 1
	mov	DPTR,#jumtab

IF BUS

        djnz    period,pwmtest          ; has period expired ? 3/4 (3 if branch not taken)
        setb    PWMOUT                  ; set PWM bit 2
;        mov     period,#RELOAD          ; reload counter 256 is OK !!!!
        mov     pwmtim,duty1            ; and PWM value 3
inter010:
        jmp     inter000                ; 3  sum = 11 istruction cycles
pwmtest:
        djnz    pwmtim,inter010         ; has PWM expired ? 3 if taken 3+4+3 if not taken
        clr     PWMOUT                  ; pulse finished, don't reload pwmtim as period has 2
        nop                             ; to expire first 
        nop                             ; sum 11 if pwmtim jump is not taken

ENDIF


inter000:
        jb      OldBit,lowhig
        jnb	PULSIN,noedge
        setb    OldBit
        sjmp    highf
noedge:
        inc     Rtimbuf                 ; no edge found
        ljmp    endint1

lowhig:
        jb	PULSIN,noedge     
        clr     OldBit
        ljmp    endint

highf:
        mov     a,Rtimbuf
        jz      faulty
        clr     C
        subb    a,#3					; was SHORT
        jc      found1
found0:
        mov     a,#3					; was OFFSET_1
        add     a,Rmode
        jmp     @a+DPTR
found1: mov     a,#0					; was OFFSET_2
        add     a,Rmode
        jmp     @a+DPTR

jumtab:
        ljmp    sync1							;mode 0
        ljmp    sync0
        ljmp    startbit11				;mode 6
        ljmp    startbit10
        ljmp    byte11						;mode 12
        ljmp    byte10
        ljmp    startbit21				;mode 18
        ljmp    startbit20
        ljmp    byte21						;mode 18
        ljmp    byte20
        ljmp    startbit31				;mode 24
        ljmp    startbit30
        ljmp    byte31						;mode 30
        ljmp    byte30
        ljmp    startbit41				;mode 36
        ljmp    startbit40
        ljmp    byte41						;mode 42
        ljmp    byte40
        ljmp    startbit51				;mode 48
        ljmp    startbit50
        ljmp    byte51						;mode 54
        ljmp    byte50
        ljmp    startbit61				;mode 60
        ljmp    startbit60
        ljmp    byte61						;mode 66
        ljmp    byte60
	ljmp	endmod0						;mode 72
	ljmp	endmod1
	ljmp	endmod0						;mode 78 errormode
	ljmp	endmod1
faulty:
        mov     Rmode,#ERRMOD           ; error encountered
        ljmp    endint

startbit11:
        ljmp    endint                  ; another sync bit
startbit10:
        mov     Rbytec,#8               ; done elsewhere ?
        ljmp    modinc                  ; ok now begin decoding


byte11: setb    C
        mov     a,byte_1
        rlc     a
        mov     byte_1,a
        ljmp    update

byte10: clr     C
        mov     a,byte_1
	rlc     a
        mov     byte_1,a
        ljmp    update

startbit21:
        mov     a,#ERRMOD
        ljmp    setmod

startbit20:
        mov     Rbytec,#8               ; rewind byte counter
        ljmp    modinc                  ; next mode, get 2nd byte

byte21: setb    C
        mov     a,byte_2
        rlc     a
        mov     byte_2,a
        ljmp    update

byte20: clr     C
        mov     a,byte_2
        rlc     a
        mov     byte_2,a
        ljmp    update

startbit31:
        mov     a,#ERRMOD
        ljmp    setmod

startbit30:
        mov     Rbytec,#8               ; rewind byte counter
        ljmp    modinc                  ; next mode, get 3rd byte

byte31: setb    C
        mov     a,byte_3
        rlc     a
        mov     byte_3,a
        ljmp    update
byte30: clr     C
        mov     a,byte_3
        rlc     a
        mov     byte_3,a
        ljmp    update

startbit41:
        mov     a,#ENDMOD               ; 3 byte message found
        ljmp    setmod

startbit40:
        mov     Rbytec,#8               ; rewind byte counter
        setb    flags1.0								; long message found
        ljmp    modinc                  ; next mode, get 4th byte

byte41: setb    C
        mov     a,byte_4
        rlc     a
        mov     byte_4,a
        ljmp    update

byte40: clr     C
        mov     a,byte_4
        rlc     a
        mov     byte_4,a
        ljmp    update

startbit51:
        mov     a,#ENDMOD               ; 4 byte message found
        ljmp    setmod

startbit50:
        mov     Rbytec,#8               ; rewind byte counter
        inc     flags1                  ; long message found
        ljmp    modinc                  ; next mode, get 5th byte

byte51: setb    C
        mov     a,byte_5
        rlc     a
        mov     byte_5,a
        ljmp    update

byte50: clr     C
        mov     a,byte_5
        rlc     a
        mov     byte_5,a
        ljmp    update

startbit61:
        mov     a,#ENDMOD               ; 4 byte message found
        ljmp    setmod

startbit60:
        mov     Rbytec,#8               ; rewind byte counter
        inc     flags1                  ; long message found
        ljmp    modinc                  ; next mode, get 5th byte

byte61: setb    C
        mov     a,byte_6
        rlc     a
        mov     byte_6,a
        ljmp    update

byte60: clr     C
        mov     a,byte_6
        rlc     a
        mov     byte_6,a
        ljmp    update

final1:
        mov     a,#ENDMOD               ; 4 byte message found
        ljmp    setmod

final0:
        mov     a,#ERRMOD               ; rewind byte counter
        inc     flags1                  ; long message found
        ljmp    modinc                  ; next mode, get 5th byte

sync1:
        djnz   Rsync,endint             ; collect sync bits
        ljmp   modinc                   ; done
                                      
sync0:                                  ; 0 start all over again
again:                   
        mov     Rmode,#0                ; clear mode and preset sync
        mov     Rsync,#SYNCP            ; counter
        ljmp    endint




; error mode and threeb, fourb, fiveb are no longer modified
endmod1:
        ljmp	endint                  ; if a 1 is found
endmod0:
        ljmp    endint

; end of jump table, remainder of interrupt are "ordinary" subroutines


setmod:
        mov     Rmode,a
        ljmp    endint

update:
        djnz	Rbytec,endint           ; decrement byte count byte 1
modinc:
        mov     a,Rmode                 ; update mode, mode = mode + n
        add     a,#MODADD 
        mov     Rmode,a
        ljmp    endint

endint:
        mov     Rtimbuf,#0                      ; Bit detected,reset counter
endint1:                                        ; endint1: bit in progress
        jnb     STOPPED, endint0                ; EMF pending ?
        jmp     housekeeping

endint0:
        jb      flags.1,revend                  ; reverse

        
        cjne    Rtimbas,#RINGI_1,endint2        ; check if EMF pending
        setb    P0.2                            ; turn off upper FET
        mov     PCA0CPH0,#0
        jmp     housekeeping
endint2:
        cjne    Rtimbas,#RINGI_2,endint3        ; check if EMF pending
	setb	AD0BUSY                         ; start ADC
        jmp     housekeeping        
endint3:
        cjne    Rtimbas,#RINGI_3,housekeeping   ; check if EMF pending
        setb    ADReady
        mov     PCA0CPH0,reload
        clr     P0.2                            ; turn upper FET back on
;>>>>rearrange Ringi_2        
        jmp    housekeeping

revend:
        cjne    Rtimbas,#RINGI_1,endint5        ; check if EMF pending
        setb    P0.3                            ; turn off upper FET
        mov     PCA0CPH1,#0
        jmp     housekeeping
endint5:
        cjne    Rtimbas,#RINGI_2,endint6        ; check if EMF pending
	setb	AD0BUSY                         ; start ADC
        jmp     housekeeping        
endint6:
        cjne    Rtimbas,#RINGI_3,housekeeping   ; check if EMF pending
        setb    ADReady
        mov     PCA0CPH1,reload
        clr     P0.3                            ; turn upper FET back on
;>>>>rearrange Ringi_2        
;        jmp    housekeeping

housekeeping:
        dec     Rtimbas
        pop     PSW
	pop	DPL
	pop	DPH
        pop     ACC
        reti
; end of interrupt routine


; decode instructions
decod:
	mov	A,byte_2
	anl	A,#11100000B
	cjne	A,#00000000B,decod1
	ljmp	deccon                  ; 000 consist + decoder reset
decod1:
	cjne	A,#00100000B,decod2
	ljmp	advanc                  ; 001 advanced
decod2:
	cjne	A,#01000000B,decod3
	ljmp	rueckw                  ; 010 reverse
decod3:
	cjne	A,#01100000B,decod4
	ljmp	vorwae                  ; 011 forward
decod4:
	cjne	A,#10000000B,decod5
	ljmp	function_group_1        ; 100 acessory 1
decod5:
	cjne	A,#10100000B,decod6
	ljmp	function_group_2        ; 101 acessory 2                  
decod6:
	cjne	A,#11000000B,decod7
	ljmp	telnew1                 ; 110 future exp ( not yet impl. )
decod7:
	cjne	A,#11100000B,decod7
        ljmp	progrm                  ; 111 programming
decod8:
        ljmp    telnew1                 ; nothing to decode


gettel:
	clr	C
	mov	A,#ERRMOD					; are we in error mode ?
	subb	A,Rmode_1					; correct register bank !!!!!
	jnz	noerror
	jmp	telnew1						; faulty, reset decoding
noerror:
	jc	gottel						; not yet ready ?
	ret
gottel:
	mov	A,byte_1					; OK, got a message
	xrl	A,byte_2					; now check for integrity
	xrl	A,byte_3
	xrl	A,byte_4
	xrl	A,byte_5
	xrl	A,byte_6
	jz	xxxx
        jmp     telnew1						; message corrupted, reset decoding
xxxx:
	nop
; at this point a valid DCC message has been found. Now do 
; address decoding.  xxxx is used as a breakpoint during test phase
	mov    failsf,#0                  ; reset fail safe counter
; triggers new decode sequence, telnew1 resets service mode
	mov	A,CV19
        rlc     a                       ; bit 7 set -> we are in a consist
        mov     CONSIST_ACTIVE,C
noconsist:
	mov	A,byte_1
	jnz	longtt						; no global command
	mov	A,#10100000B			; set functions ...
	orl	A,flags1
	mov	flags1,A
	jmp	decod							; decode broadcast

longtt:
	mov	A,byte_1
	add	A,#(-11000000B)
	jnc	constst						; check for consist
	jb	CV29.5,long_address						; 
	jmp	telnew						; long adress, but address not enabled
long_address:
	mov	A,byte_1					; check two byte address
	xrl	A,CV17
	jz	long01
        jmp     telnew1						; first byte does not match
long01:
	mov	A,byte_2					; 
	xrl	A,CV18
	jz	long02
        jmp     telnew1						; second byte does not match
long02:
	mov	byte_2,byte_3			; ok, extended address detected
	mov	byte_3,byte_4			; shuffle bytes 5->4,4->3,3->2 so structure matches 
	mov	byte_4,byte_5                   ; short address
	mov	byte_5,byte_6
				
	dec	flags1						; decrease bytecount
	mov	A,flags1					; set bit RF
	orl	A,#10000000B
	jb	CONSIST_ACTIVE,motor_only
	setb	flags1.5					; motor functions if not in consist
motor_only:
	jmp	decod							; decode long address

constst:
	jnb	CONSIST_ACTIVE,shortt; decode short
	mov	A,byte_1
	xrl	A,CV19
	anl	A,#01111111B			; first bit contains direction
	jnz	shortt						; no consist either, may be short ?
	setb	CONSIST_DECODED
	setb	flags1.5					; motor function allowed
	jmp	decod							; decode consist					

shortt:
	jb	CV29.5,servtst		; service ?
	mov	A,byte_1
IF BUS
        anl     a,#11111110B            ; when Bus active, the next odd address is valid too
ENDIF
	xrl	A,CV1
	jnz	servtst						; service ?
	setb	flags1.7					; set Bit RF
	jnb	CONSIST_ACTIVE,sh1
	setb	flags1.5					; motor functions allowed if not in consist
sh1:
	jmp	decod

servtst:
	jb      SERVMOD,ser1
	jmp	telnew1
ser1:
	mov	A,byte_1
	add	A,#0x90
	jc	ser2						; decode service mode
	jmp	telnew1						; I give up, nothing to decode
ser2:
        jmp     service

; this is the big switchboard for bits
function_group_2:
IF BUS
        jnb     byte_1.0,bus001		; even address -> normal functions
        jmp     telnew1			; odd, ignore for function group 2
bus001:
ENDIF
        mov     a,lokfun                ; get previous pattern
        swap    a                       ; F5..F8 -> bit 4..7
        mov     r0,#byte_2              ; set up ointer for xchd
        xchd    a,@r0                   ; copy lower nibble into byte_2
        swap    a                       ; get F5..F8 back to upper nibble
        mov     byte_2,a                ; and save
        sjmp    fgrou3
function_group_1:
IF BUS
        jnb     byte_1.0,bus006		; even address -> normal functions
	setb	F1
	jb	byte_2.3,bus006		; if function 1 = 1 -> rail 
	clr	F1
bus006:
ENDIF
        jb      CV29.1,light_source     ; FL set by fun4 or extra bit
        call    copy_light              ; copy light into byte_2
light_source:
        mov     a,lokfun                ; get previous pattern
        mov     r0,#byte_2              ; set up pointer for xchd
        xchd    a,@r0                   ; copy lower nibble into a
        mov     byte_2,a                ; and save
fgrou3:
        jb      CONSIST_ACTIVE,merge    ; OK need to shuffle bits
        mov     lokfun,a                ; store as is
        jmp     merend
; now check for consist...
; rest slightly complicated see line 686
        jmp     telnew1
        jmp     telnew1
merge:
        mov     a,CV21
        jb      CONSIST_DECODED,noninvert
        cpl     a                       ; complement mask if not in consist
noninvert:
        push    acc                     ; and save
        anl     a,byte_2
        mov     byte_2,a                ; save (mask & message)
        pop     acc
        cpl     a                       ; get nmask
        anl     a,lokfun                ; (nmask & old)
        orl     a,byte_2                ; (mask&message)|(nmask&old)
        mov     lokfun,a                ; and store
merend:
        call    head_light
        jmp     telnew1

head_light:
        jnb     CONSIST_ACTIVE,set_head_light
        jnb     CONSIST_DECODED,not_decoded_h
        jnb     CV22.0,jrear_light
        setb    FRONT_LIGHT
        jnb     LIGHT_CONSIST,head_off
        jmp     head_consist
not_decoded_h:
        jb      CV22.0,jrear_light
set_head_light:
        setb    FRONT_LIGHT
        jnb     LIGHT_BASE,head_off
head_consist:
        jnb     FORWARD,head_off
        jmp     jrear_light
head_off:
        clr     FRONT_LIGHT
jrear_light:
        jnb     CONSIST_Active,set_rear_light
        jnb     CONSIST_DECODED,not_decoded_r
        jnb     CV22.1,end_set_light
        setb    REAR_LIGHT
        jnb     LIGHT_CONSIST,rear_off
        jmp     rear_consist
not_decoded_r:
        jb      CV22.1,end_set_light
set_rear_light:
        setb    REAR_LIGHT
        jnb     LIGHT_BASE,rear_off
rear_consist:
        jb      FORWARD,rear_off
        jmp     end_set_light
rear_off:
        clr     REAR_LIGHT
end_set_light:

access7:
        mov     C,lokfun.0
        mov     FUN1,C
        mov     C,lokfun.1
        mov     FUN2,C
        mov     C,lokfun.2
        mov     FUN3,C
        mov     C,lokfun.3
        mov     FUN4,C
        mov     C,lokfun.4
        mov     FUN5,C
        mov     C,lokfun.5
        mov     FUN6,C
        mov     C,lokfun.6
        mov     FUN7,C
        mov     C,lokfun.7
        mov     FUN8,C

        mov     C,REAR_LIGHT
        mov     RLIGHT,C
        mov     C,FRONT_LIGHT
        mov     FLIGHT,C
        ret

IF BUS
BusServo:
	jb	F1,Rail
        mov      a,byte_2
;        jb       byte_2.5,Rail            ; Bus runs on rails
;        clr     F1
        anl      a,#00001111b             ; restrict to 4 bits
        jz       Servo1                   ; 0000 remains unaltered
        dec      a                        ; real speed is value - 1
Servo1:
        rl      a                       ; >>> integrate intermediate step
        mov     C,byte_2.4
        mov     ACC.0,C                   ; intermediate speed step
        rl      a
;        add     a,#31                   ; 45 -> 1ms -28/2 = 31
        subb    a,#101                   ; 45 -> 1ms -28/2 = 31
        cpl     a
        inc     a
        mov     duty1,a                ; load register
        jmp     telnew1

Rail:
;        setb    F1                      ; F1 set: Rail, cleared: Road
        mov     duty1,#62               ; on rail steering is blocked at neutral
        jmp     telnew1

ENDIF        



; entry for forward direction
vorwae:
IF BUS
        jb      byte_1.0, BusServo
ENDIF        
        jb      CV29.1,vorwae1          ; for 14 speed steps light is derived from speed
        call    copy_light              ; otherwise extract light information
vorwae1:
        jnb     flags1.5,vorwae2        ; function allowed ?
        jmp     telnew1

vorwae2:
        setb    flags.2                 ; set actual direction flag
        jmp     rueck1
;btfss   CV29,1
;	call    copy_light
;	btfss   flags1,5                ; function allowed ?  >>>>>>;
;	goto    telnew1
;	bcf     flags,2                 ; forward
;	goto    rueck1
;rueckw  btfss   CV29,1
;	call    copy_light
;	btfss   flags1,5                ; function allowed ?
;	goto    telnew1
;	bsf     flags,2                 ; reverse       
;rueck1  call    diradj                  ; direction affected by CV19,CV29
;	call    head_light              ; update lights etc.
;>>>8051
; entry for reverse direction
rueckw:
IF BUS
        jb      byte_1.0, BusServo
ENDIF        
        jb      CV29.1,rueckw1          ; for 14 speed steps light is derived from speed
        call    copy_light
rueckw1:
        jnb     flags1.5,rueckw2         ; function allowed ?
        jmp     telnew1

rueckw2:
        clr     flags.2                 ; set actual flag
; generic for both directions
rueck1:
        call    diradj
        mov      a,byte_2
        anl      a,#00001111b             ; restrict to 4 bits
        jz       advan4                   ; 0000 remains unaltered
        dec      a                        ; real speed is value - 1
        jnz     rueck2                    ; original 1 is emergency stop
        mov     loksoll,#0
        mov     lokist,#0               ; stop immediately
	jmp     telnew1
advan4: mov     loksoll,#0
        jmp     telnew1
rueck2:

;        clr      C
        rl       a                      ; multiply by 2 for intermediate speed step ??????
        mov      x_value,a
        mov      x_base,#28
        jb       CV29.1,s28sp           ; interpolated speed 
speedmode:                      
        jb       CV29.4,table           ; or speed table 
rueck3:
        call     interp
        call     telnew1
        ret

; adjust for intermediate 28 speed
s28sp:
	mov	a,byte_2
        jb      ACC.4,intermediate      ; intermediate speed step ?
        dec     x_value
intermediate:
        jmp      speedmode

; get speed from CV67++
table:
        mov     dptr,#CV67m
        movc    a,@a+dptr               ; pointer is still in a
        mov     loksoll,a               ; store in loksoll
        jmp     telnew1

; interpolates current speed between CV5 and CV2, no intermediate speed yet
; (CV5-CV2)*x_value/x_base
interp:
        clr      C
        mov      a,CV5
        subb     a,CV2
        jnc      inter1
;??????stope:
        mov      loksoll,#0
        ret
inter1:
        mov      B,a         ; y1
        mov      a,x_value    ; x
        mul      ab
        mov      r3,a
        mov      r2,b
        mov      a,x_base           ; base
        call     div16x8
        add      a,CV2
        mov      loksoll,a
        jmp      telnew1


; copies light information from speed/func1 to LIGHT_CONSIST
; or LIGHT_BASE according to CONSIST_DECODED
copy_light:
        jb      CONSIST_DECODED,copy_consist
        mov     C,byte_2.4
        mov     LIGHT_BASE,C
        ret
copy_consist:
        mov     C,byte_2.4
        mov     LIGHT_CONSIST,C
        ret

; decodes advanced functions
advanc:
IF BUS
        jnb     byte_1.0,bus002
        mov     a,byte_3
        anl     a,#01111111b
        jz      advanb1
        dec     a
advanb1:
        mov     b,#192                  ; 25500 when multiplied by 127
        mul     ab
        add     a,#(low 12250)          ; add 16 bit offset
        mov     steerlow,a                ; >>> scratch ?
        mov     a,b
        addc    a,#(high 12250)
        mov     steerhgh,a              ; steer contains 16bit value of PWM

        jmp     telnew1
bus002:
ENDIF
        jnb     flags1.5,advanc01       ; function allowed ?
advanc02:
        jmp     telnew1
advanc01:
        mov     a,byte_2
        cjne    a,#00111111b,advanc02   ; unknown command, as of now only 127 speed allowed
        setb    flags.2
        jb      byte_3.7,advan1         ; forward or reverse
        clr     flags.2
advan1:
        call    diradj                 ; adjust direction according to CV29 and consist
        mov     a,byte_3                ; get speed
        anl     a,#01111111b            ; restrict to 0..127
        jz      advan3                 ; zero unaltered
        dec     a
        jnz     advan3                  ; no emergency stop
        mov     loksoll,#0
        mov     lokist,#0
advan3:
        mov      x_value,a
        mov      x_base,#127
        jmp      rueck3               ; do interpolation similar to 28 speed

; adjust direction according to CV29.0 and consist
diradj:
        jnb      CV29.0,adj01         ; direction in CV29 not reversed
        cpl      flags.2
adj01:
        jnb      CONSIST_DECODED,nodir  ; not in consist
        jnb      CV19.7,nodir           ; consist does not reverse direction
        cpl      flags.2
nodir:
        ret


; decoder and consist control
deccon: 
IF BUS
        jnb     byte_1.0,bus003
        jmp     telnew1
bus003:
ENDIF
        jb      flags1.7,deccon01
        jmp     telnew1
deccon01:
        mov     a,byte_2
        jz      deccon02                ; non-zero consist ?
        jmp     conhan                  ; handle consist
deccon02:
        setb    SERVMOD                 ; enter service mode
        mov     timer2,#0                  ; reset service watch dog
stope:
        mov     lokist,#0
        mov     loksoll,#0
; PWM -> 0 >>>>>
        clr     flags.1
        clr     flags.2                 ;  default forward
        call    setpwm                  ; stop immediately
        jmp     telnew


; handle consist: address "0" deactivate, else activate. CV19.7 contains direction ?
conhan:
        mov     a,byte_2
        cjne    a,#00010010b,conhan1    ; set consist without changing direction
        cjne    a,#00010011b,conhan2    ; set consist changing direction
        jmp     telnew1
conhan2:
        mov     CV19,byte_3             ; 
        clr     CV19.7                  ; no change in direction
        jmp     contest                 ; test if consist to be deactivated
conhan1:
        mov     CV19,byte_3
        setb    CV19.7                  ; change in direction

contest:
        mov     a,CV19
        anl     a,#01111111b
        jnz     constore                ; address is "0", deactivate
        mov     CV19,#0
constore:
        mov     byte_4,CV19             ; prepare data and pointer for flash storage
        mov     r0,#CV19
        call    writatx                 ; store
        jmp     ackn
        

; program instruction decoded, not service mode e.g. programming on the main
progrm:
IF BUS
        jnb     byte_1.0,bus004
        jmp     telnew1
bus004:
ENDIF
        jnb     CONSIST_DECODED,progrm01        ; when consist address decoded, ignore programming ???
        jmp     telnew1
progrm01:
        mov     byte_6,byte_5                   ; error byte -> 6
        mov     a,byte_2
        cjne    a,#11100100b,progrm02           ; verify requested
        jmp     dverify
progrm02:
        cjne    a,#11101100b,progrm03           ; write requested
        jmp     dwrite
progrm03:
        jmp     telnew1


; what is dverify/dwrite doing in conjunction with progrm

; CVs are mirrored in CVStorage with the same offset than in RAM, so the relative pointer
; can be the same. This is necessary, as movx writes only(???) !
flash_write:

        mov     r0,#CV22        ; set read pointer
        mov     dptr,#CV22m     ; set write pointer to flash
        mov     r1,#(low lokist-low CV22); set loop counter ???? end of table ???
        clr     ea              ; disable interrupts
        mov     psctl,#03h      ; set PSEE and PSWE
        mov     flkey,#0a5h     ; write key code
	mov  	flkey,#0f1h
        clr     a
        movx    @dptr,a         ; write any byte into location to erase flash

        mov     psctl,#01h      ; set write_enable
wloop:
        mov     a,@r0           ; get byte
        mov     flkey,#0a5h     ; write key code
	mov  	flkey,#0f1h
        movx    @dptr,a         ; write byte
        inc     r0              ; update read pointer
        inc     dptr            ; update write pointer
        djnz    r1,wloop        ; loop
        setb    ea              ; reenable interrupts
        ret


; EEPROM Write/Read Routines
; data in byte_4, address in R0 (Ram)
; Address in Flash at same offset as in Ram
writatx:
        mov     dptr,#CVStorage ; set data pointer to begin of storage area
        mov     a,r0            ; get offset pointer into RAM/Flash
        movc    a,@a+dptr       ; get byte from Flash
        xrl     a,byte_4        ; if equal, no write necessary
        jz      writdone
        anl     a,byte_4        ; check if only transitions 1->0 (xrl+and=0 1->0 transitions)
        jnz     flash_write     ; no complete EEPROM has to be erased/written
        mov     a,r0            ; get offset
        add     a,DPL
        mov     DPL,a           ; and add to datapointer
        clr     ea              ; disable interrupts
        mov     a,byte_4        ; get byte to be written
        mov     flkey,#0a5h     ; write key code
	mov  	flkey,#0f1h
        movx    @dptr,a         ; write byte
        setb    ea              ; reenable interrupts
writdone:
        ret


; >>>>>> dwrite2 and dwrite3 what are they doing ?????
; generic write ??????????
dwrite1:                        ; get bytes in standardized order
        call    shiftb
dwrite:
        call    twoxor          ; second identical message ?
        jc      dwrite2
        jmp     telnew          ; not yet two conscutive messages
; entry for direct write without check for 2nd identical byte.
dwrite2:
        call    getCV           ; get address of byte in RAM
dwrite3:
        jnc     dwrite04
        jmp     telnew          ; if C set, out of range
dwrite04:
        jnz     dwrite05
        jmp     telnew          ; if zero, invalid CV
dwrite05:
;        add     a,#low CV22     ; add offset into RAM
        mov     r0,a            ; set pointer
        mov     a,byte_4        ; get byte to be stored
        cjne    r0,#low CV67,dwrite06  ; set C if r0 =< CV67 
dwrite06:
        jnc     dwrite07        ;if equal or less       ; t
        mov     @r0,a           ; store in RAM if <CV67
dwrite07:
        call    writatx         ; store in Flash        ; byte_4->data,r0->address
        jmp     ackn


; decoded if in service mode
service:
        mov    timer2,#0                ; reset timeout									; >>>>later ??
        mov     a,flags1
        anl     a,#00000011b
        jz      pagedp                  ; 3 byte message -> paged
        mov     a,byte_1
        cjne    a,#01110100b,service01
        jmp     dverify                 ; verify CV direct
service01:
        cjne    a,#01111100b,service02
        jmp     dwrite1                 ; write CV direct
service02:
        cjne    a,#01111100b,service03
        jmp     bitverify               ; bitverify
service03:
        jmp     telnew1

; paged program/verify mode
pagedp:
        mov     a,byte_1
        anl     a,#11110111b            ; mask out read/write, decoded later
        cjne    a,#01110101b,paged01
        jmp     propag
paged01:
        cjne    a,#01110100b,paged02
        jmp     proCV29                 ; manipulate CV29
paged02:
        anl     a,#11111100b
        cjne    a,#01110000b,paged03
        jmp     proreg                  ; register programming
paged03:
        jmp     telnew                  ; no valid code found


; CV29 special direct access
proCV29:
        call    shiftb1                 ; errorbyte->6, data->4
        call    twoxor
        jnc     proCV2901
        jmp     telnew
proCV2901:
        jb      byte_1.3,CV29ver        ; verify or write ?
        call    sCV29                   ; get pointer and check range
        jmp     dwrite3
CV29ver:
        mov     a,byte_4
        cjne    a,CV29,CV29ver01
        jmp     ackn                    ; OK they match
CV29ver01:
        jmp     telnew

; program/verify page register
propag:	call    shiftb1                 ; errorbyte->6, data->4
        call    twoxor
	jnb	byte_3.3,pagver         ; verify
	mov	pagreg,byte_4
	jmp	ackn
	
pagver:	mov	a,byte_4		; compare byte_4 and pagreg
	cjne	a,pagreg,pagver01
	jmp	ackn
pagver01:
	jmp	telnew			; not matched
					

; program/verify register 0..3 in paged mode
proreg:
	call	shiftb1
	call	twoxor
	jnb	CY,proreg01
	jmp	telnew
proreg01:
	mov	a,pagreg		; bits only allowed in bit 0..6 after decrement
	dec	a
	rl	a
	jnb	CY,proreg02
	jmp	telnew
proreg02:
	rl	a
	jnb	CY,proreg03
	jmp	telnew
proreg03:
	mov	scratch,a
	mov	a, byte_1
	anl	a,#00000011B
	mov	byte_3,a
	jnb	byte_1.3, dverif2
	jmp	dwrite2

; generic verify: verifies byte_4 and location in flash 
dverif1:
        call    shiftb          ; normalize bytes
dverify:
        call    twoxor          ; two identical pacvkets ?
        jb      CY,dverif2
        jmp     telnew
dverif2:
        call    getCV
        jnz     dverif3
        jmp     telnew

dverif3:
        mov     dptr,#CVStorage
        movc    a,@a+dptr
        cjne    a,byte_4,dverif4; compare bytes
        jmp     ackn            ; acknowledge
dverif4:
        jmp     telnew          ; they did not match


; verify single bit -> faster than checking 255 values...
bitverify:
        call    shiftb
        call    twoxor
        jc      bitver1
        jmp     telnew
bitver1:
        call    getCV
        jnz     bitver2
        jmp     telnew
bitver2:                        
        mov     r0,a            ;set pointer                    >>> verify flash !!!!
        mov     a,byte_4
        anl     a,#00000111B    ; strip to 0..7
        inc     a
        mov     scratch,a       ; set bit loop counter
        mov     a,@r0           ; get CV to be checked
bitloop:
        rrc     a
        djnz    scratch,bitloop ; carry contains 1 or 0 to be compared to byte_4.3      
        jc      one_set
        jnb     byte_4.3,bitackn
        jmp     telnew          ; they dont match
one_set:
        jb      byte_4.3,bitackn
        jmp     telnew          ; they dont match

bitackn:
        jmp     ackn



; normalize bytes
shiftb:
        mov     byte_6,byte_4
        mov     byte_4,byte_3
        mov     byte_3,byte_2
        ret

shiftb1:
        mov     byte_6,byte_3
        mov     byte_4,byte_2
        ret


; finds address of CV. Input: byte_3, output: address in A. CY as error flag
getcv:
        clr     c
        mov     a,byte_3
        cjne    a,#(1-1),get01
        jmp     sCV1
get01:
        cjne     a,#(2-1),get02
        jmp      sCV2
get02:
        cjne     a,#(3-1),get03
        jmp      sCV3
get03:
        cjne     a,#(4-1),get04
        jmp      sCV4
get04:
        cjne     a,#(5-1),get05
        jmp      sCV5
get05:
        cjne     a,#(8-1),get08
        jmp      sCV8
get08:
        cjne     a,#(9-1),get09
        jmp      sCV9
get09:
        cjne     a,#(17-1),get17
        jmp      sCV17
get17:
        cjne     a,#(18-1),get18
        jmp      sCV18
get18:  cjne     a,#(19-1),get19
        jmp      sCV19
get19:
        cjne     a,#(21-1),get21
        jmp      sCV21
get21:
        cjne     a,#(22-1),get22
        jmp      sCV22
get22:
        cjne     a,#(29-1),get29
        jmp      sCV29
get29:
        cjne     a,#(52-1),get52
        jmp      sCV52
get52:
        cjne     a,#(53-1),get53
        jmp      sCV53
get53:
        cjne     a,#(54-1),get54
        jmp      sCV54
get54:
        cjne     a,#(55-1),get55
        jmp      sCV55
get55:
        add      a,#(NOT(94-2))      ; upper limit
        jc       geterror          ; greater than upper limit
        clr      c
        subb     a,#((67-1)-(94-1)-3)
        jnc      geterror          ; less than lower limit
        jmp      sCV67X
geterror:
        mov      a,#0
        ret

; get address and test for high/low limits of values in CV's
;---------------------------
; not "0" and <128
sCV1:
        mov      a,byte_4
        jz       sCV1err
        add      a,#128
        cpl      c
sCV1err:
        cpl      c
        mov      a,# low CV1
        ret

;---------------------------
sCV2:
        mov      a,#low CV2
        ret
;---------------------------
sCV3:
        mov      a,#low CV3
        ret
;---------------------------
sCV4:
        mov      a,#low CV4
        ret
;---------------------------
sCV5:
        mov      a,#low CV5
        ret
;---------------------------
sCV8:
        mov      a,#low CV8
        ret

;---------------------------
; "1" not allowed
sCV9:
        mov      a,byte_4
        cjne     a,#1,sCV9OK
        cpl      C
sCV9OK: mov      a,#low CV9
        ret

;---------------------------
; must be between "192d" and "231d"
sCV17:
        mov      a,#byte_4             ; use cjne !!!!
        add      a,#11100111b
        jc       sCV17a
        add      a,#(11100111b-11000000b+2)
        jc       sCV17OK
sCV17a:
        cpl      C
sCV17OK:
        mov      a,#low CV17
        ret

;---------------------------
sCV18:
        mov      a,#low CV18
        ret

;---------------------------
sCV19:
        mov      a,#low CV19
        ret

;---------------------------
sCV21:
        mov      a,#low CV21
        ret


;---------------------------
sCV22:
        mov      a,byte_4
        anl      a,#11111100b     ; mask out valid bits
        jz       sCV22OK
        cpl      C
sCV22OK:
        mov      a,#low CV22
        ret


;---------------------------
sCV29:
        mov      a,byte_4
        anl      a,#11001100b     ; mask out valid bits
        jz       sCV29OK
        cpl      C
sCV29OK:
        mov      a,#low CV29
        ret

;---------------------------
sCV52:
        mov      a,#low CV52
        ret

;---------------------------
sCV53:
        mov      a,#low CV53
        ret

;---------------------------
sCV54:
        mov      a,#low CV54
        ret

;---------------------------
sCV55:
        mov      a,#low CV55
        ret

;---------------------------
sCV67X:
        mov      a,byte_3
        add      a,#(-(67-(low CV29)-2))
        clr      C
        ret


; checks if 2 consecutive checksums are equal sets CY if yes.
twoxor:
        mov     a,byte_6
        cjne    a,lasxor,twoxor1
        setb    CY              ; they are equal
        mov     lasxor,a
        ret
twoxor1:
        clr     CY              ; they are different
        mov     lasxor,a
        ret

; send acknowledge pulse
ackn:
        setb    FLIGHT                 ; turn front light on
	setb    RLIGHT
;  setb     flags,5                 ; enable setting of motor speed
	clr     flags.2
	clr     flags.1
        mov     lokist,#0ffh
	mov     reload,#0               ; turn motor on
;        call    EMF_c_end
	call    setpwm                  ; ??????
; delay 15..20ms
l1:
        mov     a,Rtimbas_1
        jnb     ACC.7,l1
l2:
        mov     a,Rtimbas_1
        jb      ACC.7,l2
l3:
        mov     a,Rtimbas_1
        jnb     ACC.7,l3
l4:
        mov     a,Rtimbas_1
        jb      ACC.7,l4

	clr     RLIGHT
	clr     FLIGHT                 ; turn front light off
        mov     reload,#0ffh
	mov     duty,#0            ; turn motor off
;        call    EMF_c_end
        mov     lokist,#0		; reset algorithm
;        mov     PWM_val,#0
	mov	Integ,#0

	call    setpwm                  ; ??????
        jmp     telnew


; prepare for new message 
telnew1:
        clr     SERVMOD	
;        mov     lasxor,#0       ; is this really necessary ?
; entry point for service mode, keep service mode on
telnew:	call	prepint
        ret          


; prepares all registers for decoding a new message
prepint:
	mov	byte_1,#0					; erase message bytes
	mov	byte_2,#0
	mov	byte_3,#0
	mov	byte_4,#0
	mov	byte_5,#0
	mov	byte_6,#0
	mov	flags1,#0               ; service mode ????
	mov	Rbytec_1,#8				; rewind byte counter
	mov	Rsync_1,#SYNCP			; preset SYNC counter
	mov	Rmode_1,#0					; start decoding
	ret


; all "slower" timing is done here
timact:
        mov     a,Rtimbas_1
        jb      ACC.7,flahi
        clr     flags.0
        ret
flahi:  jnb     flags.0,inct1
        ret                             ; if already set, do nothing

; every 22usec*256=5.6ms this part of the code is called
inct1:
        setb    flags.0                 ; set done bit
        djnz    failsf,nofail           ; check for timout->no DCC messages
        mov     a,loksoll
        jnz     nofail                  ; keep speed
; the idea here is to keep the speed constant, once rail contact/DCC is lost.
; if the command is "0", the loco creeps until contact is reestablished.
; later a second timeout may be needed to prevent a runaway
IF BUS
        jnb     F1,nofail               ; on rail only
ENDIF
        mov     loksoll,Creep              ; otherwise creep
nofail:
        mov     timer1,#GTAKT
        jnb     SERVMOD,tima1
        djnz    timer2,nofail1
        clr     SERVMOD
nofail1:
        ret

tima1:
        mov     a,flags
        rr      a
        xrl     a,flags
        anl     a,#00000010b            ; compare actual and nominal direction
        jnz     rwechsl
; the following code is a mess, some major housekeeping is called for
; this code checks for the difference in lokist/loksoll, if they don't
; match, load acc/dec timers and watch for high/low limits.
; setting lokist after reset should be done in begin.
        mov     a,loksoll
        cjne    a,lokist,tima02
        jmp     sgleich
tima02:
        jc      decspe                  ; decrease speed
        djnz    timacc,tima03           ; no action required yet
        inc     lokist
        mov     timacc,CV3              ; load acc. timer
        jb      flags.3,tima05          ; check if reset occurred
        mov     lokist,loksoll          ; if yes, set speed immediately
        setb    flags.3                 ; and clear flag
        jmp     braus
tima05:
        mov     a,CV3                   ; check if CV3=0
        jnz     tima06                  ; if yes, set speed immediately
        mov     lokist,loksoll
        jmp     braus
tima06:
        mov     a,CV2                   ; check if upper limit exceeded
        clr     CY                      ; >>>> use cjne !
        subb    a,lokist
        jb      CY,minspe
        mov     lokist,CV2              ; if yes, set to upper limit
minspe:
        jmp     braus

decspe:
        djnz    timacc,tima03           ; no action required yet
        dec     lokist                  ; decrease speed
        mov     timacc,CV4              ; load dec. timer
        mov     a,CV4                   ; check if CV4=0
        jnz     decspe1                 ; if yes, set speed immediately
        mov     lokist,loksoll
        jmp     braus
decspe1:
        mov     a,lokist                ; CV2 ?????????
        clr     CY
        subb    a,CV2
        jnc     decspe01
        mov     lokist,#0
decspe01:
        jmp     braus
        

rwechsl:
        mov     a,lokist
        jz      isnull
        dec     lokist                  ; already 0 ?
        mov     timacc,#BREMS2          ; reload dec. timer
        jmp     decspe1

isnull:
        mov     C,flags.2               ; copy nominal direction
        mov     flags.1,C               ; into actual direction

braus:
        call    setpwm                  ; 
sgleich:
;        mov     timacc,#1               ; prepare timacc for immediate action
tima03:
        ret


setpwm:
; >>> store last values + chsum in RAM for recovery
; loksoll/lokist/flags
; check for "0" and set flag t.b.d.
        mov     a,lokist
        jnz     setpwm1
        setb    STOPPED                 ; prevent interrupt from accessing PWM
        setb    P0.3                    ; turn off upper FET's
        setb    P0.2
	mov	PCA0CPM0,#02H           ; PWM0 off
	mov	PCA0CPM1,#02H           ; PWM1 off
;        mov     PWM_val,#0
	mov	Integ,#0
        ret
setpwm1:
;        cpl     a
;        mov     reload,a
        clr     STOPPED
        jb      flags.1,revpwm          ; reverse ?
gopwm:
        setb    P0.3                    ; turn off upper FET 1
	mov	PCA0CPM1,#02H           ; PCA 8 bit, ECOM clear, ch 1
        ret
revpwm:         
        setb    P0.2
	mov	PCA0CPM0,#02H           ; PCA 8 bit, ECOM clear, ch 1
        ret

emfcon:
        jb      ADReady,emfcon1
        ret
emfcon1:
        clr     ADReady
;        mov     a,ADC0
;        mov     duty,lokist             ; to be replaced by algorithm
; set lokist -> duty
;        mov     duty,lokist
;        ret

; the following code is translated from PIC code
; developed by Gil Fuchs/Mike Bolton

;------------------------------------
;  EMF feedback control
;
;  PWM_val  - desired value = lokist      (I[n])
;  EMF_val  - current EMF sample value   (F[n])
;  PWM_thr  - PWM control signal  threshold (C[n])
;
;  Implemented a PDFF (PI+) controller
; C(z) = Kp *[ (I(z) - F(z))(Ki * T * Z) / (Z - 1)  + Kfr * I(z) - F(z) ]
; C[n] = C[n-1] + Kp * [Ki(I[n] - F[n]) + Kfr * (I[n] - I[n-1]) - (F[n] - F[n-1])]
;    this can be rearranged:
; C[n] = C[n-1] + Kp * [Ki(I[n] - F[n]) + (Kfr * I[n] - F[n]) - (Kfr * I[n-1] - F[n-1])]
; define Fval[n] := - (Kfr*I[n] - F[n]) = (F[n] - Kfr * I[n])



EMF_control:
        mov     a,lokist
;        mov     a,PWM_val
	clr	F0              ; Flag for positive = 0
	clr	C               ; in preparation for subb
        subb    a,ADC0          ; subtract
;        subb    a,ADC          ; subtract
	jnc	EMF_c_8         ; positive ?
	setb	F0              ; set negative Flag
	cpl     a               ; change sign
        inc     a
EMF_c_8:
        inc     a               ; rounding error ?
        mov     B,a
        mov     a,EMF_Ki
        mul     ab
        push    ACC             ; save for later
        mov     a,B
        anl     a,#0f0h         ; is it going to overflow
        jnz     EMF_c_11
        pop     ACC
        swap    a
        anl     a,#0fh
        mov     r7,a            ; save for later use
        mov     a,B             ; get high byte
        swap    a
        anl     a,#0f0h
        orl     a,R7
        jmp     EMF_c_12
EMF_c_11:
        pop     ACC             ; restore stack
        mov     a,#0ffh         ; set to max
EMF_c_12:
        jnb     F0,EMF_c_9	; check for underflow
        clr     C
        xch     a,Integ
        subb    a,Integ
        jnc     EMF_c_10
        mov     a,#0
        jmp     EMF_c_10

EMF_c_9:
        add     a,Integ		; check for overflow
        jnc     EMF_c_10
        mov     a,#0ffh         ; max limit
EMF_c_10:
        mov     Integ,a

;        jmp     EMF_c_end       ; >>>>>>>>>>>>>>>>>>>

; here the integral part is done
; no threshold yet

; here the strange stuff

        mov     a,lokist               ; assume no phase out needed 
;        mov     B,EMF_Kfr		; new md 12/2005
	mov	B,lokist		; new md 12/2005
	jmp	EMF_c_7c		; new md 12/2005
        cjne    a,EMF_Kfr,EMF_c_7e
        jmp     EMF_c_7d                ; OK they are equal
EMF_c_7e:
        jc      EMF_c_7d                ; phase out
        mov     a,lokist		; redundant ???
EMF_c_7d:				; 
        mul     ab 			; 
; end of strange stuff
EMF_c_7c:
        clr     C
        mov     a,B
        subb    a,ADC0            ; subtract
;        subb    a,ADC            ; subtract
        jnc     EMF_c_3
	setb	F0              ; set negative Flag
	cpl     a               ; change sign
        inc     a
EMF_c_3:
        mov     B,a
        mov     a,EMF_Kp
        mul     ab
        push    ACC             ; save for later
        mov     a,B
        anl     a,#0f0h         ; is it going to overflow
        jnz     EMF_c_4
        pop     ACC
        swap    a
        anl     a,#0fh
        mov     r7,a            ; save for later use
        mov     a,B             ; get high byte
        swap    a
        anl     a,#0f0h
        orl     a,R7
        jmp     EMF_c_5
EMF_c_4:
        pop     ACC             ; restore stack
        mov     a,#0ffh         ; set to max
EMF_c_5:
        jnb      F0,EMF_c_6      ; was jnb ????

        clr     C
        subb    a,Integ
        cpl     a
        inc     a
        jc      EMF_c_end
        mov     a,#0
        jmp     EMF_c_end

EMF_c_6:
        add     a,Integ
        jnc     EMF_c_end
        mov     a,#0ffh         ; max limnit
EMF_c_end:
        nop
        mov     duty,a
;        mov     duty,Integ
        cpl     a
        mov     reload,a
        ret

IF BUS
SteerPWM:
        jnb     CCF2,endsteer           ; has match already occurred ?
        clr     CCF2                    ; prepare for next match
        mov     PCA0CPL2,steerlow       ; reload PWM >>> later timered !!!
        mov     PCA0CPH2,steerhgh
endsteer:
        ret
ENDIF


; sets all registers/peripherals
begin:
        anl     PCA0MD, #NOT(040h)      ; clear Watchdog Enable bit
   	mov     PSW,#00001000B          ; use register bank 1
	mov	R0,#0
	mov	R1,#0
	mov	R2,#0
	mov	R3,#0
	mov	R4,#0
	mov	R5,#0
	mov	R6,#0
	mov	R7,#0
   	mov     PSW,#00000000B          ; use register bank 0


IF BUS
        ; set 16bit PWM and crossbar t.b.d.
ENDIF
            ; Enable the Port I/O Crossbar
        mov     XBR0,#00011111b         ; force PWM0 and PWM1 to P0.5/P0.6
                                          ; a
				
        mov	XBR1,#80h               ; PWM0 sel. 80 for 0/1
        mov	XBR2,#40h               ; weak pullups,enabled Xbar enabled

        mov	P0MDOUT,#11101110b      ; make outputs push-pull except A/D(4) and Pulse 0.0
        mov	P0MDIN,#11101111b       ; make P0.4 analog input

;        mov     REF0CN,#02H            ; internal Vref with VDD
        mov     REF0CN,#00AH            ; internal Vref with VDD
	mov	PCA0MD,#08h             ; PWM sysclock on PCA0					; System Clock, no WD
	mov	PCA0CPM0,#02H           ; PCA 8 bit, ECOM cleared, channel 0!					; 8 bit PWM
	mov	PCA0CN,#040h            ; PCA timer enabled						; enable PCA Timer

	mov	PCA0CPM1,#02H           ; PCA 8 bit, ECOM cleared, channel 1					; 8 bit PWM
IF BUS
        mov     PCA0CPM2, #82H          ; ECOM to be set ????
        mov     PCA0CPL2,#(low 24500)
        mov     PCA0CPH2,#(high 24500)  ; preload with neutral
ENDIF
	mov	OSCICN,#00000111b
;        mov     OSCICL,#FREQUENCY       ; tuning value for non-300s

	mov	AMX0SL,#84h             ; singlr ended, P0.4 ADC						; single ended, channel 6
	mov	ADC0CF,#011h            ; sysclock/4 gain 						; gain = 1, clock = ?
;	mov	ADC0CF,#010h            ; sysclock/4 gain 0.5						; gain = 0.5, clock = ?
	mov	ADC0CN,#80h             ; ADC enabled triggered by ADC0BUSY						; ADC on
	mov	ADC0LT,#0FFh            ; Window <= not uwsed
	mov	ADC0GT,#00h             ;  Window >= not used						; window function not needed	
				; Initialize LED to OFF
;	clr			GREEN_LED
	mov	R5,#02h
	mov	R6,#02h
	mov	R7,#03h
;           mov SP, #STACK_TOP                  ; init stack pointer to end of allocated RAM
        mov     SP, #80h                     ; init stack pointer to end of allocated RAM
        mov     TL0,#-44
        mov     TH0,#-44
        mov     CKCON,#0                ; System clock/12
        mov     TMOD,#00000010b         ; Timer 0 Mode 2 8 bit auto relaod
        mov     TCON,#00010000b         ; Timer 0 on
        mov     IE,#10000010b           ; Interrupt enable, timer 0, all


        mov     r0,#CV22        ; set write pointer
        mov     a,#CV22
        mov     dptr,#01C00h    ; set read pointer
        add     a,dpl           ; add offset
        mov     dpl,a           
        mov     r1,#(low lokist-low CV22); set loop counter

bloop:
        clr     a
        movc    a,@a+dptr       ; get byte from Code segment
        mov     @r0,a           ; and store
        inc     r0
        inc     dptr
        djnz    r1,bloop        ; loop





; temporary >>>>>>
;        mov     CV1,#3                  ; address
;        mov     CV2,#0                  ; min speed
;        mov     CV3,#25                 ; acceleration
;        mov     CV4,#25                 ; decelaration
;        mov     CV5,#33                 ; top speed
;        mov     CV17,#0                 ; long address 1
;        mov     CV18,#0                 ; long address 2
;        mov     CV19,#0                 ; consist
;        mov     CV29,#0                 ; config
        mov     PCA0CPH0,#0FFh
        mov     duty,#0h                      ; start with speed 0
IF BUS
        mov     duty1,#31
        mov     period,#1                       ; ensure clean start of PWM
        clr     PWMOUT
ENDIF
        mov     loksoll,#0
        mov     lokist,#0
        mov     timacc,#1
        mov     reload,#0FFh
        mov     Integ,#0
        mov     failsf,#0
        clr     flags1.1
        mov     flags,#0
        setb    P0.7                            ; just in case for PULsIN
        setb    STOPPED
        setb    P0.2                            ; turn both upper FET's off
        setb    P0.3
;        mov     PWM_val,#0
	mov	Integ,#0
;        mov     EMF_Ki,#080h
;IF BUS
;        mov     EMF_Ki,#02h
;        mov     EMF_Kp,#60h
;        mov     EMF_Kfr,#166
;ELSE
;        mov     EMF_Ki,#10h
;        mov     EMF_Kp,#050h
;        mov     EMF_Kfr,#166
;ENDIF
        call    prepint                         ; reset registers to start decoding

loop:
        lcall   gettel                           ; check for valid message, decode etc.
IF BUS
        lcall   SteerPWM
ENDIF
        lcall   timact
        lcall   emfcon
        sjmp    loop






;   Divide 16 bit unsigned value by 8 bit unsigned value
;   Target: 8051 processor family
;
;  Copyright 1996 by Thad Smith, Boulder Colorado, USA
;  Permission is hereby granted to use this code for any purpose
;  without compensation provided that this copyright notice is
;  kept with the code.  It has been tested but is not warranteed.
;
; On entry: r2, r3 = dividend msby,lsby, respectively
;           A      = divisor
; On exit:  A      = quotient
;           r3     = remainder
;           C      = overflow if set
;
; Registers used: A, r2, r3, C
; Execution time: 110 - 118 cycles, including call and return
;
div16x8:
        cpl     a               ; get 2's comp of divisor,
        add     a,#1            ; .. clear C (or set if divisor 0)
        xch     a,r2            ; save divisor, get msby dividend
        mov     r4,#9           ; set loop counter
div1:
        jnc     div2            ; dividend high order 1 bit?
        add     a,r2            ; --yes, it's bigger than divisor
        setb    c               ; quotient bit = 1
        sjmp    div3
div2:
        add     a,r2            ; subtract divisor
        jc      div3            ; if pos store 1 bit
        subb    a,r2            ; neg, restore a, clr C
div3:   xch     a,r3            ; get lsby dividend & quotient
        rlc     a               ; save quotient bit, shift lsby dividend
        xch     a,r3
        rlc     a               ; shift msby dividend
        djnz    r4,div1         ; if not finished

        rrc     a               ; C = overflow, restore remainder
        xch     a,r3            ; A=quotient, r3 = remainder
        ret

        CSEG AT (CVStorage + 02Ah)

;Offs:           DB      47      ; offset of RAM to mirror same structure
CV22m:	        DB	1
CV29m:	        DB	2	; config 28 speed steps
CV17m:   	DB	0
CV18m:   	DB	0
CV19m:   	DB	0
CV21m:   	DB	1
; end of bit-adressable section
IF BUS
CV1m:    	DB	06h     ; address
ELSE
CV1m:    	DB	03h     ; address
ENDIF
CV2m:    	DB	0       ; starting voltage
CV3m:    	DB	20
CV4m:    	DB	20
CV5m:    	DB	100     ; Vmax
CV8m:    	DB	1
CV9m:    	DB	1
CV52m:   	DB	200     ;Kp
CV53m:   	DB	20      ;Ki
CV54m:   	DB	0       ;KFr
CV55m:   	DB	20      ;Creep Speed

CV67m:          DB      0,9,18,27
                DB      36,45,54,63
                DB      70,79,88,96
                DB      105,114,123,132
                DB      141,150,159,168
                DB      177,186,197,206
                DB      215,224,233,243
                

END
